
#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include <Debug.h>
#include <TranslatorAddOn.h>
#include <TranslatorFormats.h>
#include <DataIO.h>
#include <Message.h>
#include <String.h>
#include <Screen.h>
#include <Bitmap.h>


const uint32 k8BitHexDump = '8BHX';
const uint32 k24BitHexDump = '24HX';
const uint32 k32BitHexDump = '32HX';

inline void
ThrowError(status_t error, const char *debugPrint = 0)
{
	if (error != B_OK) {
		if (debugPrint)
			printf("%s %s\n", debugPrint, strerror(error));
		throw error;
	}
}

inline void
ThrowIf(status_t error, bool condition, const char *debugPrint = 0)
{
	if (condition)
		ThrowError(error, debugPrint);
}

template <class InitCheckable>
void
InitCheck(InitCheckable *checkThis)
{
	ThrowError(checkThis->InitCheck());
}

inline size_t
ThrowOrSize(status_t result, const char *debugPrint = 0)
{
	if (result < B_OK)
		ThrowError(result, debugPrint);
	
	return result;
}

static status_t
BitmapToHexDump(BPositionIO *input, BPositionIO *output,
	const translator_info *, uint32 outType)
{
	BScreen screen(B_MAIN_SCREEN_ID);
	char *buffer = 0;
	try {
		TranslatorBitmap bitmapHeader;
		if (input->Read(&bitmapHeader, sizeof(bitmapHeader)) != sizeof(bitmapHeader))
			return B_ERROR;
	
		int32 width = (int32)(B_BENDIAN_TO_HOST_FLOAT(bitmapHeader.bounds.right)
			- B_BENDIAN_TO_HOST_FLOAT(bitmapHeader.bounds.left)) + 1;
		int32 height = (int32)(B_BENDIAN_TO_HOST_FLOAT(bitmapHeader.bounds.bottom)
			- B_BENDIAN_TO_HOST_FLOAT(bitmapHeader.bounds.top)) + 1;

		int32 rowBytes = B_BENDIAN_TO_HOST_INT32(bitmapHeader.rowBytes);
		color_space space = (color_space)B_BENDIAN_TO_HOST_INT32(bitmapHeader.colors);

		const int32 kMaxColumnWidth = 80;
		int32 columnWidth = kMaxColumnWidth;
		int32 pixelWidth = 0;
		color_space targetSpace = B_COLOR_8_BIT;
		
		BString outString;

		switch (outType) {
			case k8BitHexDump:
				pixelWidth = sizeof("0x00,");
				targetSpace = B_COLOR_8_BIT;
				outString 	<< "const color_space kBitmapColorSpace = B_COLOR_8_BIT;\n";
				break;
			case k24BitHexDump:
				pixelWidth = sizeof("0x00,0x00,0x00,");
				targetSpace = B_RGB24;
				outString 	<< "const color_space kBitmapColorSpace = B_RGB24;\n";
				break;
			case k32BitHexDump:
				pixelWidth = sizeof("0x00,0x00,0x00,0x00,");
				targetSpace = B_RGB_32_BIT;
				outString 	<< "const color_space kBitmapColorSpace = B_RGB_32_BIT;\n";
				break;
		}

		BBitmap *targetBitmap = new BBitmap(BRect(0, 0, width - 1, height - 1), targetSpace);
		outString << "const int32 kBitmapBytesPerRow = " << targetBitmap->BytesPerRow() << ";\n";
		delete targetBitmap;
		outString << "const int32 kBitmapWidth = " << width << ";\n"
				  << "const int32 kBitmapHeight = " << height << ";\n\n";
		
		outString << "const unsigned char kBitmapBits [] = {";

		ThrowOrSize( output->Write(outString.String(), outString.Length()) , "initial write");

		size_t readSize = rowBytes;
		
		buffer = new char[readSize];
		off_t pos = input->Position();

		columnWidth /= pixelWidth;
		columnWidth *= pixelWidth;

		int32 column = 0;

		for (int index = 0; index <= height-1; index++) {
			ThrowOrSize( input->Seek(pos + (off_t)index * rowBytes, SEEK_SET) , "seek error");
			ThrowIf(B_ERROR, ThrowOrSize(
				input->Read(buffer, readSize),
				"read error") != readSize, "read error, too little returned");

			const char *end = buffer + readSize;
			rgb_color pixel;
			pixel.alpha = 255;

			// XXX: this assumes that pixels in all input formats are 4 bytes.
			for (const char *src = buffer; src < end; src += 4) {
				switch (space) {
					case B_RGBA32:
						pixel.alpha = src[3];
						pixel.red = src[2];
						pixel.green = src[1];
						pixel.blue = src[0];
						break;
					case B_RGB32:
						pixel.red = src[2];
						pixel.green = src[1];
						pixel.blue = src[0];
						break;
					case B_RGBA32_BIG:
						pixel.alpha = src[3];
						pixel.red = src[2];
						pixel.green = src[1];
						pixel.blue = src[0];
						break;
					case B_RGB32_BIG:
						pixel.red = src[2];
						pixel.green = src[1];
						pixel.blue = src[0];
						break;
					default:
						break;
				}
				outString = "";
				if (column == 0)
					outString << "\n\t";
				
				switch (outType) {
					case k8BitHexDump:
						{
							uchar beColorPaletteColor = screen.IndexForColor(pixel);
							char tmp[8];
							sprintf(tmp, "0x%02x,", beColorPaletteColor);
							outString << tmp;
						}
						break;
					case k24BitHexDump:
						{
							char tmp[16];
							sprintf(tmp, "0x%02x,0x%02x,0x%02x,", pixel.blue, pixel.green, pixel.red);
							outString << tmp;
						}
						break;
					case k32BitHexDump:
						{
							char tmp[24];
							sprintf(tmp, "0x%02x,0x%02x,0x%02x,0x%02x,", pixel.blue, pixel.green, pixel.red, pixel.alpha);
							outString << tmp;
						}
						break;
				}

				ThrowOrSize( output->Write(outString.String(), outString.Length()));

				column += pixelWidth;
				if (column >= columnWidth)
					column = 0;
			}
		}
		outString = "";
		outString << "\n};\n\n";
		output->Seek(output->Position() - 1, SEEK_SET); // don't print the last comma
		ThrowOrSize( output->Write(outString.String(), outString.Length()) );
		
	} catch (status_t error) {
	}

	delete [] buffer;
	return B_OK;
}

static status_t
HexDump8ToHexDump8(BPositionIO *, BPositionIO *, const translator_info *)
{
	return B_OK;
}

static bool
CanHandle(uint32 inType, uint32 outType)
{
	switch (inType) {
		case B_TRANSLATOR_BITMAP:
			return outType == 0
				|| outType == k8BitHexDump
				|| outType == k24BitHexDump
				|| outType == k32BitHexDump;
	}
	
	return false;
}

static bool
GuessType(BPositionIO *inData, uint32 &type)
{
	char bytes[16];

	if (inData->Read(&bytes, sizeof(bytes)) != sizeof(bytes)) 
		return false;

	if ((*(TranslatorBitmap *)&bytes).magic
			== B_HOST_TO_BENDIAN_INT32(B_TRANSLATOR_BITMAP)) {
		type = B_TRANSLATOR_BITMAP;
		return true;
	}
	return false;
}

status_t
Identify(BPositionIO *inData, const translation_format *inFormat,
	BMessage *, translator_info *outInfo, uint32 outType)
{
	uint32 inType;
	bool guessed = false;
	if (inFormat) 
		inType = inFormat->type;
	else if (GuessType(inData, inType))
		guessed = true;
	else
		return B_NO_TRANSLATOR;

	/* make sure we can handle the type */
	if (!CanHandle(inType, outType))
		return B_NO_TRANSLATOR;

	// should check header here
	
	outInfo->type = B_TRANSLATOR_BITMAP;
	
	outInfo->type = inType;
	int32 i = 0;
	while (inputFormats[i].type != 0) {
		if (inputFormats[i].type == outInfo->type)
			break;
		i++;
	}

	outInfo->translator	= 0;
	outInfo->group = inputFormats[i].group;
	outInfo->quality = inputFormats[i].quality;
	outInfo->capability	= inputFormats[i].capability;
	strcpy(outInfo->name, inputFormats[i].name);
	strcpy(outInfo->MIME, inputFormats[i].MIME);

	return B_OK;
}

status_t
Translate(BPositionIO *inData, const translator_info *inInfo,
	BMessage *, uint32 outType, BPositionIO *outData)
{
	if (!inInfo || !CanHandle(inInfo->type, outType))
		return B_NO_TRANSLATOR;

	if (outType == 0) {
//		if (inInfo->type == k8BitHexDump)
//			outType = B_TRANSLATOR_BITMAP;
//		else
			outType = k8BitHexDump;
	}
	
	if (inInfo->type == k8BitHexDump && outType == B_TRANSLATOR_BITMAP)
		; // return HexDump8ToBitmap(inData, outData, inInfo);
	else if (inInfo->type == B_TRANSLATOR_BITMAP
		&& (outType == k8BitHexDump || outType == k24BitHexDump || outType == k32BitHexDump))
		return BitmapToHexDump(inData, outData, inInfo, outType);
	else
		;
	return B_OK;
}

int
main()
{
}

char translatorName[]	= "Hex Dump Bitmap Translator";
char translatorInfo[]	= "Version 1.0, " __DATE__;
int32 translatorVersion	= 100;

translation_format inputFormats[] = {
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 1.0, 1.0,
		"image/x-be-bitmap", "Be Bitmap" },
	{ 0, 0, 0, 0, "", "" }
};
	
translation_format outputFormats[] = {
	{ k8BitHexDump, B_TRANSLATOR_BITMAP, 1.0, 1.0,
		"text/source-code", "8-bit bitmap hex dump" },
	{ k24BitHexDump, B_TRANSLATOR_BITMAP, 1.0, 1.0,
		"text/source-code", "24-bit bitmap hex dump" },
	{ k32BitHexDump, B_TRANSLATOR_BITMAP, 1.0, 1.0,
		"text/source-code", "32-bit bitmap hex dump" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 1.0, 1.0,
		"image/x-be-bitmap", "Be Bitmap" },
	{ 0, 0, 0, 0, "", "" }
};
